from typing import List
from dataclasses import dataclass
from abc import ABC, abstractmethod

@dataclass
class Chunk:
    """Dataclass representing a text chunk with metadata."""
    text: str
    start_index: int
    end_index: int
    token_count: int

class BaseChunker(ABC):
    """Abstract base class for all chunker implementations.
    
    All chunker implementations should inherit from this class and implement
    the chunk() method according to their specific chunking strategy.
    """
    def __init__(self, tokenizer):
        """Initialize the chunker with a tokenizer.
        
        Args:
            tokenizer: Tokenizer object to be used for tokenizing text
        """
        self.tokenizer = tokenizer
        self._tokenizer_backend = self._get_tokenizer_backend()

    def _get_tokenizer_backend(self):
        """Return the backend tokenizer object."""
        if "transformers" in str(type(self.tokenizer)):
            return "transformers"
        elif "tokenizers" in str(type(self.tokenizer)):
            return "tokenizers"
        elif "tiktoken" in str(type(self.tokenizer)):
            return "tiktoken"
        else:
            raise ValueError("Tokenizer backend not supported")
    
    def _encode(self, text: str):
        """Encode text using the backend tokenizer."""
        if self._tokenizer_backend == "transformers":
            return self.tokenizer.encode(text)
        elif self._tokenizer_backend == "tokenizers":
            return self.tokenizer.encode(text).ids
        elif self._tokenizer_backend == "tiktoken":
            return self.tokenizer.encode(text)
        else:
            raise ValueError("Tokenizer backend not supported.")
    
    def _encode_batch(self, texts: List[str]):
        """Encode a batch of texts using the backend tokenizer."""
        if self._tokenizer_backend == "transformers":
            return self.tokenizer.batch_encode_plus(texts)['input_ids']
        elif self._tokenizer_backend == "tokenizers":
            return self.tokenizer.encode_batch(texts)
        elif self._tokenizer_backend == "tiktoken":
            return self.tokenizer.encode_batch(texts)
        else:
            raise ValueError("Tokenizer backend not supported.")

    @abstractmethod
    def chunk(self, text: str) -> List[Chunk]:
        """Split text into chunks according to the implementation strategy.
        
        Args:
            text: Input text to be chunked
            
        Returns:
            List of Chunk objects containing the chunked text and metadata
        """
        pass

    def __call__(self, text: str) -> List[Chunk]:
        """Make the chunker callable directly.
        
        Args:
            text: Input text to be chunked
            
        Returns:
            List of Chunk objects containing the chunked text and metadata
        """
        return self.chunk(text)
    
    def __repr__(self) -> str:
        """Return string representation of the chunker."""
        return f"{self.__class__.__name__}()"


from dataclasses import dataclass
from typing import List, Optional
import numpy as np
import re
import importlib.util
from tokenizers import Tokenizer

from .base import BaseChunker
from .sentence import Sentence, SentenceChunk

@dataclass
class SemanticSentence(Sentence): 
    text: str
    start_index: int
    end_index: int
    token_count: int
    embedding: Optional[np.ndarray] = None

@dataclass
class SemanticChunk(SentenceChunk):
    text: str
    start_index: int
    end_index: int
    token_count: int
    sentences: List[SemanticSentence] = None
            
class SemanticChunker(BaseChunker):
    def __init__(
        self,
        tokenizer: Tokenizer,
        sentence_transformer_model: str,
        similarity_threshold: Optional[float] = None,
        similarity_percentile: Optional[float] = None,
        max_chunk_size: int = 512,
        initial_sentences: int = 1,
        sentence_mode: str = "heuristic",
        spacy_model: str = "en_core_web_sm"
    ):
        """Initialize the SemanticChunker.

        Args:
            tokenizer: Tokenizer for counting tokens
            sentence_transformer_model: Name of the sentence-transformers model to load
            max_chunk_size: Maximum tokens allowed per chunk
            similarity_threshold: Absolute threshold for semantic similarity (0-1)
            similarity_percentile: Percentile threshold for similarity (0-100)
            initial_sentences: Number of sentences to start each chunk with
            sentence_mode: "heuristic" or "spacy" for sentence splitting
            spacy_model: Name of spaCy model to use if sentence_mode="spacy"

        Raises:
            ValueError: If parameters are invalid
            ImportError: If required dependencies aren't installed
        """
        super().__init__(tokenizer)

        if max_chunk_size <= 0:
            raise ValueError("max_chunk_size must be positive")
        if similarity_threshold is not None and (similarity_threshold < 0 or similarity_threshold > 1):
            raise ValueError("similarity_threshold must be between 0 and 1")
        if similarity_percentile is not None and (similarity_percentile < 0 or similarity_percentile > 100):
            raise ValueError("similarity_percentile must be between 0 and 100")
        if similarity_threshold is not None and similarity_percentile is not None:
            raise ValueError("Cannot specify both similarity_threshold and similarity_percentile")
        if similarity_threshold is None and similarity_percentile is None:
            raise ValueError("Must specify either similarity_threshold or similarity_percentile")
        if sentence_mode not in ["heuristic", "spacy"]:
            raise ValueError("sentence_mode must be 'heuristic' or 'spacy'")

        self.max_chunk_size = max_chunk_size
        self.similarity_threshold = similarity_threshold
        self.similarity_percentile = similarity_percentile
        self.initial_sentences = initial_sentences
        self.sentence_mode = sentence_mode

        # Initialize sentence transformer
        if not importlib.util.find_spec("sentence_transformers"):
            raise ImportError(
                "sentence-transformers is not installed. "
                "Install it with 'pip install sentence-transformers'"
            )
        try:
            from sentence_transformers import SentenceTransformer
            self.sentence_transformer = SentenceTransformer(sentence_transformer_model)
        except Exception as e:
            raise ImportError(
                f"Failed to load sentence-transformers model '{sentence_transformer_model}'. "
                f"Error: {str(e)}"
            ) from e

        # Initialize spaCy if explicitly requested
        self.nlp = None
        if sentence_mode == "spacy":
            if not importlib.util.find_spec("spacy"):
                raise ImportError(
                    "spaCy is not installed. Install it with 'pip install spacy' "
                    "and download the model with 'python -m spacy download en_core_web_sm', "
                    "or use sentence_mode='heuristic' instead."
                )
            try:
                import spacy
                self.nlp = spacy.load(spacy_model)
            except OSError as e:
                raise ImportError(
                    f"spaCy model '{spacy_model}' not found. "
                    f"Download it with 'python -m spacy download {spacy_model}' "
                    "or use sentence_mode='heuristic' instead."
                ) from e

    def _split_sentences_spacy(self, text: str) -> List[str]:
        """Split text into sentences using spaCy."""
        doc = self.nlp(text)
        return [str(sent).strip() for sent in doc.sents if str(sent).strip()]

    def _split_sentences_heuristic(self, text: str) -> List[str]:
        """Split text into sentences using rule-based approach."""
        pattern = r'(?<=[.!?])\s+(?=[A-Z])|(?<=[.!?])(?=\s*[A-Z])|(?<=[.!?])\s*$'
        sentences = re.split(pattern, text)
        return [s.strip() for s in sentences if s.strip()]

    def _split_sentences(self, text: str) -> List[str]:
        """Split text into sentences using specified mode."""
        if self.sentence_mode == "heuristic":
            return self._split_sentences_heuristic(text)
        else:
            return self._split_sentences_spacy(text)

    def _compute_similarity_threshold(self, all_similarities: List[float]) -> float:
        """Compute similarity threshold based on percentile if specified."""
        if self.similarity_threshold is not None:
            return self.similarity_threshold
        else:
            return float(np.percentile(all_similarities, self.similarity_percentile))
    
    def _prepare_sentences(self, text: str) -> List[Sentence]:
        """Prepare sentences with precomputed information.
        
        Args:
            text: Input text to be processed
            
        Returns:
            List of Sentence objects with precomputed token counts and embeddings
        """
        if not text.strip():
            return []
            
        # Split text into sentences
        raw_sentences = self._split_sentences(text)
        
        # Get start and end indices for each sentence
        sentence_indices = []
        current_idx = 0
        for sent in raw_sentences:
            start_idx = text.find(sent, current_idx)
            end_idx = start_idx + len(sent)
            sentence_indices.append((start_idx, end_idx))
            current_idx = end_idx

        # Batch compute embeddings for all sentences
        embeddings = self.sentence_transformer.encode(raw_sentences, convert_to_numpy=True)
        
        # Batch compute token counts
        token_counts = [len(encoding) for encoding in self._encode_batch(raw_sentences)]
        
        # Create Sentence objects with all precomputed information
        sentences = [
            SemanticSentence(
                text=sent,
                start_index=start_idx,
                end_index=end_idx,
                token_count=count,
                embedding=embedding
            )
            for sent, (start_idx, end_idx), count, embedding 
            in zip(raw_sentences, sentence_indices, token_counts, embeddings)
        ]
        
        return sentences

    def _get_semantic_similarity(self, embedding1: np.ndarray, embedding2: np.ndarray) -> float:
        """Compute cosine similarity between two embeddings."""
        similarity = self.sentence_transformer.similarity(embedding1, embedding2)
        return similarity

    def _compute_group_embedding(self, sentences: List[Sentence]) -> np.ndarray:
        """Compute mean embedding for a group of sentences."""
        return np.mean([sent.embedding for sent in sentences], axis=0)

    def _group_sentences(self, sentences: List[Sentence]) -> List[List[Sentence]]:
        """Group sentences based on semantic similarity, ignoring token count limits.
        
        Args:
            sentences: List of Sentence objects with precomputed embeddings
            
        Returns:
            List of sentence groups, where each group is semantically coherent
        """
        if len(sentences) <= self.initial_sentences:
            return [sentences]
            
        # Get or compute similarity threshold
        if self.similarity_percentile is not None:
            # Compute all pairwise similarities
            all_similarities = [
                self._get_semantic_similarity(
                    sentences[i].embedding,
                    sentences[i+1].embedding
                )
                for i in range(len(sentences)-1)
            ]
            similarity_threshold = float(np.percentile(all_similarities, self.similarity_percentile))
        else:
            similarity_threshold = self.similarity_threshold

        groups = []
        current_group = sentences[:self.initial_sentences]
        current_embedding = self._compute_group_embedding(current_group)
        
        for sentence in sentences[self.initial_sentences:]:
            # Compare new sentence against mean embedding of entire current group
            similarity = self._get_semantic_similarity(
                current_embedding,
                sentence.embedding
            )
            
            if similarity >= similarity_threshold:
                # Add to current group
                current_group.append(sentence)
                # Update mean embedding
                current_embedding = self._compute_group_embedding(current_group)
            else:
                # Start new group
                if current_group:
                    groups.append(current_group)
                current_group = [sentence]
                current_embedding = sentence.embedding
        
        # Add final group
        if current_group:
            groups.append(current_group)
            
        return groups

    def _create_chunk(self, 
                     sentences: List[Sentence], 
                     similarity_scores: List[float] = None) -> SemanticChunk:
        """Create a chunk from a list of sentences."""
        if not sentences:
            raise ValueError("Cannot create chunk from empty sentence list")
            
        # Compute chunk text and token count from sentences
        text = " ".join(sent.text for sent in sentences)
        token_count = sum(sent.token_count for sent in sentences) + (len(sentences) - 1)  # Add spaces
        
        return SemanticChunk(
            text=text,
            start_index=sentences[0].start_index,
            end_index=sentences[-1].end_index,
            token_count=token_count,
            sentences=sentences
        )

    def _split_chunks(self, sentence_groups: List[List[Sentence]]) -> List[SemanticChunk]:
        """Split sentence groups into chunks that respect max_chunk_size.
        
        Args:
            sentence_groups: List of semantically coherent sentence groups
            
        Returns:
            List of SemanticChunk objects
        """
        chunks = []
        
        for group in sentence_groups:
            current_chunk_sentences = []
            current_tokens = 0
            
            for sentence in group:
                test_tokens = current_tokens + sentence.token_count + (1 if current_chunk_sentences else 0)
                
                if test_tokens <= self.max_chunk_size:
                    # Add to current chunk
                    current_chunk_sentences.append(sentence)
                    current_tokens = test_tokens
                else:
                    # Create chunk if we have sentences
                    if current_chunk_sentences:
                        chunks.append(self._create_chunk(current_chunk_sentences))
                    
                    # Start new chunk with current sentence
                    current_chunk_sentences = [sentence]
                    current_tokens = sentence.token_count
            
            # Create final chunk for this group
            if current_chunk_sentences:
                chunks.append(self._create_chunk(current_chunk_sentences))
        
        return chunks

    def chunk(self, text: str) -> List[SemanticChunk]:
        """Split text into semantically coherent chunks using two-pass approach.
        
        First groups sentences by semantic similarity, then splits groups to respect
        max_chunk_size while maintaining sentence boundaries.

        Args:
            text: Input text to be chunked

        Returns:
            List of SemanticChunk objects containing the chunked text and metadata
        """
        if not text.strip():
            return []

        # Prepare sentences with precomputed information
        sentences = self._prepare_sentences(text)
        if len(sentences) < self.initial_sentences:
            return [self._create_chunk(sentences)]

        # First pass: Group sentences by semantic similarity
        sentence_groups = self._group_sentences(sentences)
        
        # Second pass: Split groups into size-appropriate chunks
        chunks = self._split_chunks(sentence_groups)
        
        return chunks

    def __repr__(self) -> str:
        threshold_info = (
            f"similarity_threshold={self.similarity_threshold}"
            if self.similarity_threshold is not None
            else f"similarity_percentile={self.similarity_percentile}"
        )
        return (f"SemanticChunker(max_chunk_size={self.max_chunk_size}, "
                f"{threshold_info}, "
                f"initial_sentences={self.initial_sentences}, "
                f"sentence_mode='{self.sentence_mode}')")


from typing import List
from tokenizers import Tokenizer

from .base import Chunk, BaseChunker
class TokenChunker(BaseChunker):
    def __init__(self, tokenizer: Tokenizer, chunk_size: int = 512, chunk_overlap: int = 128):
        """Initialize the TokenChunker with configuration parameters.

        Args:
            tokenizer: The tokenizer instance to use for encoding/decoding
            chunk_size: Maximum number of tokens per chunk
            chunk_overlap: Number of tokens to overlap between chunks

        Raises:
            ValueError: If chunk_size <= 0 or chunk_overlap >= chunk_size
        """
        super().__init__(tokenizer)
        if chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")

        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def chunk(self, text: str) -> List[Chunk]:
        """Split text into overlapping chunks of specified token size.

        Args:
            text: Input text to be chunked

        Returns:
            List of Chunk objects containing the chunked text and metadata
        """
        if not text.strip():
            return []

        # Encode full text
        encoding = self._encode(text)
        text_tokens = encoding
        chunks = []

        # Calculate chunk positions
        start_indices = range(0, len(text_tokens), self.chunk_size - self.chunk_overlap)

        for start_idx in start_indices:
            # Get token indices for this chunk
            end_idx = min(start_idx + self.chunk_size, len(text_tokens))

            # Extract and decode tokens for this chunk
            chunk_tokens = text_tokens[start_idx:end_idx]
            chunk_text = self.tokenizer.decode(chunk_tokens)

            chunks.append(Chunk(
                text=chunk_text,
                start_index=start_idx,
                end_index=end_idx,
                token_count=len(chunk_tokens)
            ))

            # Break if we've reached the end of the text
            if end_idx == len(text_tokens):
                break

        return chunks

    def __repr__(self) -> str:
        return (f"TokenChunker(chunk_size={self.chunk_size}, "
                f"chunk_overlap={self.chunk_overlap})")


from typing import List, Optional
import re
from dataclasses import dataclass
from tokenizers import Tokenizer
from .base import Chunk, BaseChunker
import importlib.util
import warnings

# Check if spacy is available
SPACY_AVAILABLE = importlib.util.find_spec("spacy") is not None
if SPACY_AVAILABLE:
    try:
        import spacy
        from spacy.language import Language
    except ImportError:
        SPACY_AVAILABLE = False
        warnings.warn("Failed to import spacy despite it being installed. Using heuristic mode only.")


#TODO: Currently the SentenceChunker does not populate the sentences field in the SentenceChunk object.
# We need to implement the logic to populate the sentences field in the SentenceChunk object.

@dataclass
class Sentence: 
    text: str
    start_index: int
    end_index: int
    token_count: int

@dataclass
class SentenceChunk(Chunk):
    text: str
    start_index: int
    end_index: int
    token_count: int
    sentences: List[Sentence] = None

class SentenceChunker(BaseChunker):
    def __init__(
        self,
        tokenizer: Tokenizer,
        chunk_size: int = 512,
        chunk_overlap: int = 128,
        mode: str = "simple",
        min_sentences_per_chunk: int = 1,
        spacy_model: str = "en_core_web_sm"
    ):
        """Initialize the SentenceChunker with configuration parameters.

        Args:
            tokenizer: The tokenizer instance to use for encoding/decoding
            chunk_size: Maximum number of tokens per chunk
            chunk_overlap: Number of tokens to overlap between chunks
            mode: Sentence detection mode - "heuristic" (rule-based) or "spacy" (ML-based)
            min_sentences_per_chunk: Minimum number of sentences per chunk (defaults to 1)
            spacy_model: Name of spaCy model to use (defaults to "en_core_web_sm")

        Raises:
            ValueError: If parameters are invalid
            Warning: If spacy mode is requested but spacy is not available
        """
        super().__init__(tokenizer)

        if chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")
        if mode not in ["simple", "spacy"]:
            raise ValueError("mode must be either 'simple' or 'spacy'")
        if min_sentences_per_chunk < 1:
            raise ValueError("min_sentences_per_chunk must be at least 1")

        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.min_sentences_per_chunk = min_sentences_per_chunk

        # Initialize mode and spaCy
        self.nlp: Optional[Language] = None
        if mode == "spacy":
            if not SPACY_AVAILABLE:
                warnings.warn(
                    "Spacy not found in environment. To use spacy mode, install it using:\n"
                    "pip install spacy\n"
                    "python -m spacy download en_core_web_sm\n"
                    "Falling back to simple mode."
                )
                self.mode = "simple"
            else:
                try:
                    self.nlp = spacy.load(spacy_model)
                    # Optimize for sentence segmentation only
                    self.nlp.select_pipes(enable=["senter"])
                    self.mode = "spacy"
                except OSError:
                    warnings.warn(
                        f"Spacy model '{spacy_model}' not found. To use spacy mode, "
                        f"install it using: python -m spacy download {spacy_model}\n"
                        "Falling back to simple mode."
                    )
                    self.mode = "simple"
        else:
            self.mode = "simple"

    def _split_into_sentences_via_spacy(self, text: str) -> List[str]:
        """Split text into sentences via spaCy.

        Args:
            text: Input text to be split into sentences

        Returns:
            List of sentences
        """
        # Use spaCy's sentence segmentation
        doc = self.nlp(text)
        sentences = [sent.text.strip() for sent in doc.sents if sent.text.strip()]  
        token_counts = self._get_token_counts(sentences)
        current_pos = 0
        for sent_text, token_count in zip(sentences, token_counts):
            if not sent_text:
                continue
            start_idx = text.find(sent_text, current_pos)
            end_idx = start_idx + len(sent_text)
            current_pos = end_idx
            
            # Get the token count for the sentence
            sentences.append(
                Sentence(text=sent_text,
                         start_index=start_idx,
                         end_index=end_idx,
                         token_count=token_count)
            )
        return sentences
    
    def _split_into_sentences_simple(self, text: str) -> List[str]:
        # Fallback to heuristic mode
        # Simple rule-based sentence splitting with common abbreviations
        text = re.sub(r'([.!?])([^"])', r'\1\n\2', text)  # Add newlines after sentence endings
        text = re.sub(r'([.!?]")(\s*[A-Z])', r'\1\n\2', text)  # Handle quotes

        # Handle common abbreviations
        abbrevs = r'(?:Mr|Mrs|Dr|Prof|Sr|Jr|vs|etc|viz|al|Gen|Col|Fig|e\.g|i\.e)\.'
        text = re.sub(f'{abbrevs}\n', f'{abbrevs} ', text)

        # Handle initials and acronyms
        text = re.sub(r'([A-Z]\.[A-Z]\.)\n', r'\1 ', text)
        text = re.sub(r'([A-Z]\.[A-Z]\.)\n', r'\1 ', text)  # Run twice for consecutive initials

        # Handle decimal numbers and ellipsis
        text = re.sub(r'(\d+)\.\n(\d+)', r'\1.\2', text)  # Decimal numbers
        text = re.sub(r'\.{3}\n', '... ', text)  # Ellipsis

        sents = [s.strip() for s in text.split('\n') if s.strip()]
        token_counts = self._get_token_counts(sents)

        sentences = []
        current_pos = 0
        for sent, token_count in zip(sents, token_counts):
            start_idx = text.find(sent, current_pos)
            end_idx = start_idx + len(sent)
            current_pos = end_idx

            # Get the token count for the sentence
            sentences.append(
                Sentence(text=sent,
                         start_index=start_idx,
                         end_index=end_idx,
                         token_count=token_count)
            )
        return sentences 

    def _split_into_sentences(self, text: str) -> List[str]:
        """Split text into sentences based on the selected mode.

        Args:
            text: Input text to be split into sentences

        Returns:
            List of sentences
        """
        if self.mode == "spacy" and self.nlp is not None:
            return self._split_into_sentences_via_spacy(text)
        elif self.mode == "simple":
            return self._split_into_sentences_simple(text)
    
    def _get_token_counts(self, sentences: List[str]) -> List[int]:
        """Get token counts for a list of sentences in batch.

        Args:
            sentences: List of sentences

        Returns:
            List of token counts for each sentence
        """
        # Batch encode all sentences at once
        encoded_sentences = self._encode_batch(sentences)
        return [len(encoded) for encoded in encoded_sentences]

    def _create_chunk(
        self,
        sentences: List[Sentence],
        start_idx: int,
        token_count: int
    ) -> Chunk:
        """Create a chunk from a list of sentences.

        Args:
            sentences: List of sentences to create chunk from
            start_idx: Starting index in original text
            token_count: Total token count for the chunk

        Returns:
            Chunk object
        """
        chunk_text = " ".join([sentence.text for sentence in sentences])
        return SentenceChunk(
            text=chunk_text,
            start_index=start_idx,
            end_index=start_idx + token_count,
            token_count=token_count, 
            sentences=sentences
        )

    def chunk(self, text: str) -> List[Chunk]:
        """Split text into overlapping chunks based on sentences while respecting token limits.

        Args:
            text: Input text to be chunked

        Returns:
            List of Chunk objects containing the chunked text and metadata
        """
        if not text.strip():
            return []

        sentences = self._split_into_sentences(text)
        token_counts = [sentence.token_count for sentence in sentences]
        
        chunks = []
        current_sentences = []
        current_tokens = 0
        last_chunk_end = 0

        for i, (sentence, token_count) in enumerate(zip(sentences, token_counts)):
            # Calculate total tokens if we add this sentence
            test_tokens = current_tokens + token_count + (1 if current_sentences else 0)  # Add 1 for space between sentences
            
            can_add_sentence = (
                test_tokens <= self.chunk_size or
                (len(current_sentences) < self.min_sentences_per_chunk and len(current_sentences) + 1 <= self.min_sentences_per_chunk)
            )

            if can_add_sentence:
                # Sentence fits within limits, add it
                current_sentences.append(sentence)
                current_tokens = test_tokens
            else:
                # Sentence would exceed limits, create chunk if we have enough sentences
                if len(current_sentences) >= self.min_sentences_per_chunk:
                    chunk = self._create_chunk(
                        current_sentences,
                        last_chunk_end,
                        current_tokens
                    )
                    chunks.append(chunk)

                    # Calculate overlap for next chunk
                    if self.chunk_overlap > 0:
                        # Keep sentences from the end of current chunk until we hit overlap limit
                        overlap_sentences = []
                        overlap_tokens = 0
                        for sent, tokens in zip(reversed(current_sentences), reversed(token_counts[i-len(current_sentences):i])):
                            test_overlap_tokens = overlap_tokens + tokens + (1 if overlap_sentences else 0)
                            if test_overlap_tokens <= self.chunk_overlap:
                                overlap_sentences.insert(0, sent)
                                overlap_tokens = test_overlap_tokens
                            else:
                                break

                        current_sentences = overlap_sentences
                        current_tokens = overlap_tokens
                    else:
                        current_sentences = []
                        current_tokens = 0

                    last_chunk_end += current_tokens
                
                # Add current sentence (either after creating chunk or when forced to meet minimum)
                current_sentences.append(sentence)
                current_tokens = current_tokens + token_count + (1 if len(current_sentences) > 1 else 0)

        # Handle remaining sentences
        if current_sentences:
            chunk = self._create_chunk(
                current_sentences,
                last_chunk_end,
                current_tokens
            )
            chunks.append(chunk)

        return chunks

    def __repr__(self) -> str:
        return (f"SentenceChunker(chunk_size={self.chunk_size}, "
                f"chunk_overlap={self.chunk_overlap}, mode='{self.mode}', "
                f"min_sentences_per_chunk={self.min_sentences_per_chunk})")


from .chunker import (
    BaseChunker,
    TokenChunker,
    WordChunker,
    SentenceChunker,
    SemanticChunker,
    SPDMChunker,
    Chunk, 
    SentenceChunk,
    SemanticChunk,
    Sentence,
    SemanticSentence
)

__version__ = "0.0.2"
__name__ = "chonkie"
__author__ = "Bhavnick Minhas"

__all__ = [
    "__name__",
    "__version__",
    "__author__",
    "Sentence",
    "SemanticSentence", 
    "Chunk",
    "SentenceChunk",
    "SemanticChunk",
    "BaseChunker",
    "TokenChunker", 
    "WordChunker",
    "SentenceChunker",
    "SemanticChunker",
    "SPDMChunker"
]


from typing import List, Tuple
import re
from tokenizers import Tokenizer
from .base import Chunk, BaseChunker

class WordChunker(BaseChunker):
    def __init__(self, tokenizer: Tokenizer, chunk_size: int = 512, chunk_overlap: int = 128, mode: str = "simple"):
        """Initialize the WordChunker with configuration parameters.

        Args:
            tokenizer: The tokenizer instance to use for encoding/decoding
            chunk_size: Maximum number of tokens per chunk
            chunk_overlap: Maximum number of tokens to overlap between chunks
            mode: Tokenization mode - "heuristic" (space-based) or "advanced" (handles punctuation)

        Raises:
            ValueError: If chunk_size <= 0 or chunk_overlap >= chunk_size or invalid mode
        """
        super().__init__(tokenizer)

        if chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")
        if mode not in ["simple", "advanced"]:
            raise ValueError("mode must be either 'heuristic' or 'advanced'")

        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.mode = mode

    def _split_into_words(self, text: str) -> List[str]:
        """Split text into words based on the selected mode.
        
        Args:
            text: Input text to be split into words
            
        Returns:
            List of words
        """
        if self.mode == "simple":
            # Simple space-based splitting
            return text.split()
        elif self.mode == "advanced":
            # Advanced tokenization handling various cases
            
            # Replace common abbreviations with placeholders to preserve periods
            abbreviations = {
                r'(?i)mr\.': 'MR_ABBR',
                r'(?i)mrs\.': 'MRS_ABBR',
                r'(?i)ms\.': 'MS_ABBR',
                r'(?i)dr\.': 'DR_ABBR',
                r'(?i)prof\.': 'PROF_ABBR',
                r'(?i)sr\.': 'SR_ABBR',
                r'(?i)jr\.': 'JR_ABBR',
                r'(?i)vs\.': 'VS_ABBR',
                r'(?i)etc\.': 'ETC_ABBR',
                r'(?i)i\.e\.': 'IE_ABBR',
                r'(?i)e\.g\.': 'EG_ABBR',
                # Add numbers with decimal points
                r'(\d+)\.(\d+)': r'\1_DECIMAL_\2'
            }
            
            processed_text = text
            for pattern, replacement in abbreviations.items():
                processed_text = re.sub(pattern, replacement, processed_text)
            
            # Split on word boundaries while preserving all punctuation
            # This regex handles:
            # - Words (\w+)
            # - Contractions (word's, don't, etc.)
            # - Hyphenated words (self-aware)
            # - Various punctuation marks
            words = re.findall(
                r"\b[A-Za-z0-9]+(?:[-'][A-Za-z0-9]+)*\b"  # Words with possible hyphens/apostrophes
                r"|[.,!?;:\"'(){}[\]]+"                    # Punctuation groups
                r"|\.{3}"                                  # Ellipsis
                r"|--+"                                    # Em/En dashes
                r"|'[A-Za-z]+"                            # Starting contractions like 'tis
                r"|[A-Za-z]+'"                            # Ending contractions like ol'
                r"|[@#$%&*+=/<>~^]",                      # Special characters
                processed_text
            )
            
            # Process the words
            final_words = []
            skip_next = False
            
            for i, word in enumerate(words):
                if skip_next:
                    skip_next = False
                    continue
                    
                # Restore abbreviations
                for abbr, placeholder in abbreviations.items():
                    if placeholder in word:
                        word = re.sub(r'_ABBR$', '.', word)
                        word = re.sub(r'_DECIMAL_', '.', word)
                
                # Handle punctuation
                if word in ".,!?;:\"'(){}[]":
                    if final_words:  # Attach punctuation to previous word
                        # Don't attach if previous word already ends with punctuation
                        if final_words[-1][-1] not in ".,!?;:\"'(){}[]":
                            final_words[-1] += word
                        else:
                            final_words.append(word)
                    else:
                        final_words.append(word)
                
                # Handle ellipsis
                elif word == "...":
                    if final_words:
                        final_words[-1] += word
                    else:
                        final_words.append(word)
                
                # Handle dashes between words
                elif word.startswith("-") and i > 0:
                    if i < len(words) - 1 and re.match(r'\w+', words[i+1]):
                        final_words[-1] += word + words[i+1]
                        skip_next = True
                    else:
                        final_words.append(word)
                
                # Regular words
                else:
                    final_words.append(word)
            
            return final_words
        else:
            raise ValueError("mode must be either 'heuristic' or 'advanced'")

    def _get_token_count(self, text: str) -> int:
        """Get the number of tokens in a text string.

        Args:
            text: Input text

        Returns:
            Number of tokens
        """
        return len(self._encode(text))

    def _create_chunk(self, words: List[str], start_idx: int, end_idx: int) -> Tuple[Chunk, int]:
        """Create a chunk from a list of words.

        Args:
            words: List of words to create chunk from
            start_idx: Starting index in original text
            end_idx: Ending index in original text

        Returns:
            Tuple of (Chunk object, number of tokens in chunk)
        """
        chunk_text = " ".join(words)
        token_count = self._get_token_count(chunk_text)
        return Chunk(
            text=chunk_text,
            start_index=start_idx,
            end_index=end_idx,
            token_count=token_count
        )

    def _get_word_list_token_counts(self, words: List[str]) -> List[int]:
        """Get the number of tokens for each word in a list.

        Args:
            words: List of words

        Returns:
            List of token counts for each word
        """
        encodings = self._encode_batch(words)
        return [len(encoding) for encoding in encodings]

    def chunk(self, text: str) -> List[Chunk]:
        """Split text into overlapping chunks based on words while respecting token limits.

        Args:
            text: Input text to be chunked

        Returns:
            List of Chunk objects containing the chunked text and metadata
        """
        if not text.strip():
            return []

        words = self._split_into_words(text)
        lengths = self._get_word_list_token_counts(words)
        chunks = []
        
        # Saving the current chunk 
        current_chunk = []
        current_chunk_length = 0

        for i, (word, length) in enumerate(zip(words, lengths)):
            if current_chunk_length + length <= self.chunk_size:
                current_chunk.append(word)
                current_chunk_length += length
            else:
                chunk = self._create_chunk(
                    current_chunk,
                    i - len(current_chunk),
                    i - 1
                )
                chunks.append(chunk)

                # update the current_chunk and previous chunk
                previous_chunk_length = current_chunk_length
                
                current_chunk = []
                current_chunk_length = 0
                
                overlap = []
                overlap_length = 0
                # calculate the overlap from the current chunk in reverse
                for j in range(0, previous_chunk_length):
                  cwi = i - 1 - j
                  oword = words[cwi]
                  olength = lengths[cwi]
                  if overlap_length + olength <= self.chunk_overlap:
                    overlap.append(oword)
                    overlap_length += olength
                  else:
                    break

                current_chunk = [w for w in reversed(overlap)]
                current_chunk_length = overlap_length
                
                current_chunk.append(word)
                current_chunk_length += length
            
        # Add the final chunk if it has any words
        if current_chunk:
            chunk = self._create_chunk(
            current_chunk,
            len(words) - len(current_chunk),
            len(words) - 1
            )
            chunks.append(chunk)
        return chunks

    def __repr__(self) -> str:
        return (f"WordChunker(chunk_size={self.chunk_size}, "
                f"chunk_overlap={self.chunk_overlap}, mode='{self.mode}')")


from .base import Chunk, BaseChunker
from .token import TokenChunker
from .word import WordChunker
from .sentence import Sentence, SentenceChunk, SentenceChunker
from .semantic import SemanticSentence, SemanticChunk, SemanticChunker
from .spdm import SPDMChunker


__all__ = [
    "Chunk", 
    "BaseChunker",
    "TokenChunker",
    "WordChunker",
    "Sentence",
    "SentenceChunk",
    "SentenceChunker",
    "SemanticSentence",
    "SemanticChunk",
    "SemanticChunker",
    "SPDMChunker"
]


from typing import List
from .semantic import SemanticChunker, SemanticChunk, Sentence

class SPDMChunker(SemanticChunker):
    def __init__(
        self,
        tokenizer,
        sentence_transformer_model: str,
        similarity_threshold: float = None,
        similarity_percentile: float = None,
        max_chunk_size: int = 512,
        initial_sentences: int = 1,
        sentence_mode: str = "heuristic",
        spacy_model: str = "en_core_web_sm",
        skip_window: int = 1  # How many chunks to skip when looking for similarities
    ):
        """Initialize the SPDMChunker.
        
        Args:
            Same as SemanticChunker, plus:
            skip_window: Number of chunks to skip when looking for similarities
        """
        super().__init__(
            tokenizer=tokenizer,
            sentence_transformer_model=sentence_transformer_model,
            max_chunk_size=max_chunk_size,
            similarity_threshold=similarity_threshold,
            similarity_percentile=similarity_percentile,
            initial_sentences=initial_sentences,
            sentence_mode=sentence_mode,
            spacy_model=spacy_model
        )
        self.skip_window = skip_window

    def _merge_groups(self, groups: List[List[Sentence]]) -> List[Sentence]:
        """Merge the groups together"""
        merged_group = []
        for group in groups:
            merged_group.extend(group)
        return merged_group

    def _skip_and_merge(self,
                        groups: List[List[Sentence]],
                        similarity_threshold: float) -> List[List[Sentence]]:
        """Merge similar groups considering skip window."""
        if len(groups) <= 1:
            return groups

        merged_groups = []
        embeddings = [self._compute_group_embedding(group) for group in groups]
        
        while groups:
            if len(groups) == 1:
                merged_groups.append(groups[0])
                break
                
            # Calculate skip index ensuring it's valid
            skip_index = min(self.skip_window + 1, len(groups) - 1)
            
            # Compare current group with skipped group
            similarity = self._get_semantic_similarity(embeddings[0], embeddings[skip_index])
            
            if similarity >= similarity_threshold:
                # Merge groups from 0 to skip_index (inclusive)
                merged = self._merge_groups(groups[:skip_index + 1])
                
                # Remove the merged groups
                for _ in range(skip_index + 1):
                    groups.pop(0)
                    embeddings.pop(0)
                    
                # Add merged group back at the start
                groups.insert(0, merged)
                embeddings.insert(0, self._compute_group_embedding(merged))
            else:
                # No merge possible, move first group to results
                merged_groups.append(groups.pop(0))
                embeddings.pop(0)
        
        return merged_groups

    def chunk(self, text: str) -> List[SemanticChunk]:
        """Split text into chunks using the SPDM approach.
        
        Args:
            text: Input text to be chunked
            
        Returns:
            List of SemanticChunk objects
        """
        if not text.strip():
            return []

        # Prepare sentences with precomputed information
        sentences = self._prepare_sentences(text)
        if len(sentences) < self.initial_sentences:
            return [self._create_chunk(sentences)]

        # First pass: Group sentences by semantic similarity
        initial_groups = self._group_sentences(sentences)

        # Second pass: Merge similar groups with skip window
        merged_groups = self._skip_and_merge(initial_groups, self.similarity_threshold)
        
        # Final pass: Split into size-appropriate chunks
        chunks = self._split_chunks(merged_groups)
        
        return chunks

    def __repr__(self) -> str:
        threshold_info = (
            f"similarity_threshold={self.similarity_threshold}"
            if self.similarity_threshold is not None
            else f"similarity_percentile={self.similarity_percentile}"
        )
        return (f"SPDMChunker(max_chunk_size={self.max_chunk_size}, "
                f"{threshold_info}, "
                f"initial_sentences={self.initial_sentences}, "
                f"sentence_mode='{self.sentence_mode}', "
                f"skip_window={self.skip_window})")